import json
import logging
import re
from collections import namedtuple
from multiprocessing import Queue
from typing import Tuple, List, Optional, Union

from litellm import moderation
from pydantic import BaseModel, Field

from scholarqa.llms.litellm_helper import llm_completion
from scholarqa.llms.constants import CompletionResult
from scholarqa.llms.prompts import QUERY_DECOMPOSER_PROMPT

logger = logging.getLogger(__name__)

LLMProcessedQuery = namedtuple("LLMProcessedQuery",
                               ["rewritten_query", "keyword_query", "search_filters"])


class DecomposedQuery(BaseModel):
    earliest_search_year: str = Field(description="The earliest year to search for papers")
    latest_search_year: str = Field(description="The latest year to search for papers")
    venues: str = Field(description="Comma separated list of venues to search for papers")
    authors: Union[List[str]|str] = Field(description="List of authors to search for papers", default=[])
    field_of_study: str = Field(description="Comma separated list of field of study to search for papers")
    rewritten_query: str = Field(description="The rewritten simplified query")
    rewritten_query_for_keyword_search: str = Field(description="The rewritten query for keyword search")


def moderation_api(text: str) -> bool:
    response = moderation(text, model="omni-moderation-latest")
    return response.results[0].flagged


def validate(query: str) -> None:
    # self.update_task_state(task_id, "Validating the query")
    logger.info(
        "Checking query for malicious content with moderation api..."
    )
    try:
        if moderation_api(query):
            raise Exception(
                "The input query contains harmful content."
            )
    except Exception as e:
        logger.error(f"Query validation failed, {e}")
        raise e
    logger.info(f"{query} is valid")


def decompose_query(query: str, decomposer_llm_model: str, **llm_kwargs) -> Tuple[LLMProcessedQuery, CompletionResult]:
    search_filters = dict()
    decomp_query_res = None
    try:
        # decompose query to get llm re-written and keyword query with filters
        decomp_query_res = llm_completion(user_prompt=query, system_prompt=QUERY_DECOMPOSER_PROMPT,
                                          model=decomposer_llm_model, response_format=DecomposedQuery, **llm_kwargs)
        decomposed_query = json.loads(decomp_query_res.content)
        decomposed_query = {k: str(v) if type(v) == int else v for k, v in decomposed_query.items()}
        decomposed_query = DecomposedQuery(**decomposed_query)
        logger.info(f"Decomposed query: {decomposed_query}")
        rewritten_query, keyword_query = decomposed_query.rewritten_query, decomposed_query.rewritten_query_for_keyword_search
        if decomposed_query.earliest_search_year or decomposed_query.latest_search_year:
            search_filters["year"] = f"{decomposed_query.earliest_search_year}-{decomposed_query.latest_search_year}"
        if decomposed_query.venues:
            search_filters["venue"] = decomposed_query.venues
        if decomposed_query.field_of_study:
            search_filters["fieldsOfStudy"] = decomposed_query.field_of_study
    except Exception as e:
        logger.error(f"Error while decomposing query: {e}")
        rewritten_query = query
        keyword_query = ""
        decomp_query_res = CompletionResult(content="", model=f"error-{decomposer_llm_model}",
                                      cost=0.0,
                                      input_tokens=0,
                                      output_tokens=0, total_tokens=0,
                                      reasoning_tokens=0)

    return LLMProcessedQuery(rewritten_query=rewritten_query, keyword_query=keyword_query,
                             search_filters=search_filters), decomp_query_res



import re
from typing import Optional, Dict, Any, List
from scholarqa.utils import make_int
from langsmith import traceable
import logging
from anyascii import anyascii

logger = logging.getLogger(__name__)


def find_tldr_super_token(text: str) -> Optional[str]:
    # First, find the first instance of any token that has text "tldr" or "TLDR" in it, considering word boundaries
    tldr_token = re.search(r"\b\w*tldr\w*\b", text, re.IGNORECASE)

    if tldr_token:
        tldr_token = tldr_token.group(0)
        # Now find the word or token that the tldr_token is a subtoken in. This includes punctuation and markdown symbols
        tldr_super_token_pattern = re.compile(rf"[^\s]*{re.escape(tldr_token)}[^\s]*", re.IGNORECASE)
        tldr_super_token = re.search(tldr_super_token_pattern, text)

        if tldr_super_token:
            return tldr_super_token.group(0)
        else:
            return None
    else:
        return None


def get_section_text(gen_text: str) -> Dict[str, Any]:
    # Assume each section starts with 'Section X' followed by 'TLDR:'
    # find the first instance of any token surrounded by spaces or newlines that has text "tldr" or "TLDR" in it
    tldr_token = find_tldr_super_token(gen_text)
    curr_section = dict()
    if tldr_token is not None:
        parts = gen_text.split(tldr_token)
    else:
        parts = [gen_text]
    try:
        if len(parts) > 1:
            title = parts[0].strip()
            title = re.sub(r"\s*\(list\)", "", title)
            title = re.sub(r"\s*\(synthesis\)", "", title)
            curr_section["title"] = title.strip('#').strip()
            if tldr_token is not None:
                text_parts = parts[1].strip().split("\n", 1)
                tldr = text_parts[0]  # Assume TLDR is a single line
                text = text_parts[1]
                curr_section["tldr"] = tldr.strip('#').strip()
            else:
                text = parts[1].strip()
            curr_section["text"] = text
        else:
            raise Exception("Invalid content generated for the query by the LLM")
    except Exception as e:
        logger.exception(f"Error while parsing llm gen text: {gen_text} - {e}")
        raise e

    return curr_section


def resolve_ref_id(ref_str, ref_corpus_id, citation_ids):
    # in case of multiple papers from same author in the same year, add a count suffix
    if ref_str not in citation_ids:
        citation_ids[ref_str] = dict()
    if ref_corpus_id not in citation_ids[ref_str]:
        if citation_ids[ref_str]:
            rfsplits = ref_str.split(",")
            # in case of 2 (Doe et al., 2024), the one found later becomes (Doe et al._1, 2024) and so on...
            if len(rfsplits) > 1:
                ref_str_id = f"{rfsplits[0]}_{len(citation_ids[ref_str])},{rfsplits[1]}"
            else:
                ref_str_id = f"{ref_str}_{len(citation_ids[ref_str])}"
        else:
            ref_str_id = ref_str
        citation_ids[ref_str][ref_corpus_id] = ref_str_id
    else:
        ref_str_id = citation_ids[ref_str][ref_corpus_id]
    return ref_str_id


def pop_ref_data(ref_str_id, ref_corpus_id, fixed_quote, curr_paper_metadata) -> Dict[str, Any]:
    curr_ref = dict()
    curr_ref["id"] = ref_str_id
    curr_ref["snippets"] = [fq.strip() for fq in fixed_quote.split("...")]
    curr_ref["paper"] = dict()
    curr_ref["paper"]["corpus_id"] = make_int(ref_corpus_id)
    if curr_paper_metadata:
        #Commenting out the open access check as we switch to s2 api for the open access logic

        # if not (curr_paper_metadata.get("isOpenAccess") and curr_paper_metadata.get("openAccessPdf")):
        #     if curr_paper_metadata.get("abstract"):
        #         curr_ref["snippets"] = [s for s in curr_ref["snippets"] if
        #                                 s[:100] not in curr_paper_metadata["abstract"]]
        #     if not curr_ref["snippets"]:
        #         curr_ref["snippets"] = ["Please click on the paper title to read the abstract on Semantic Scholar."]

        curr_ref["score"] = curr_paper_metadata.get("relevance_judgement", 0)
        curr_ref["paper"]["title"] = curr_paper_metadata["title"]
        curr_ref["paper"]["authors"] = curr_paper_metadata["authors"]
        curr_ref["paper"]["year"] = make_int(curr_paper_metadata.get("year", 0))
        curr_ref["paper"]["venue"] = curr_paper_metadata["venue"]
        curr_ref["paper"]["n_citations"] = curr_paper_metadata["citationCount"]
    return curr_ref


@traceable(name="Postprocessing: Converted LLM generated output to json summary")
def get_json_summary(llm_model: str, summary_sections: List[str], summary_quotes: Dict[str, Any],
                     paper_metadata: Dict[str, Any], citation_ids: Dict[str, Dict[int, str]],
                     inline_tags=False) -> List[Dict[str, Any]]:
    text_ref_format = '<Paper corpusId="{corpus_id}" paperTitle="{ref_str}" isShortName></Paper>'
    sections = []
    llm_name_parts = llm_model.split("/", maxsplit=1)
    llm_ref_format = f'<Model name="{llm_name_parts[0].capitalize()}" version="{llm_name_parts[1]}">'
    summary_quotes = {anyascii(k): v for k, v in summary_quotes.items()}
    inline_citation_quotes = {anyascii(k): v for incite in summary_quotes.values() for k, v in
                              incite["inline_citations"].items()}
    for sec in summary_sections:
        curr_section = get_section_text(sec)
        text = curr_section["text"]
        if curr_section:
            pattern = r"(?:; )?(\d+ \| [A-Za-z. ]+ \| \d+ \| Citations: \d+)"
            replacement = r"] [\1"
            text = re.sub(pattern, replacement, text)
            text = re.sub(r"\[\]", "", text)
            curr_section["text"] = text.replace("[LLM MEMORY | 2024]", llm_ref_format)
            refs_list = []
            # tool tips inserted via span tags
            references = re.findall(r"\[.*?\]", text)
            refs_done = set()

            for ref in references:
                ref = anyascii(ref)
                if ref in summary_quotes or ref in inline_citation_quotes:
                    ref_parts = ref[1:-1].split(" | ")
                    ref_corpus_id, ref_str = ref_parts[0], f"({ref_parts[1]}, {make_int(ref_parts[2])})".replace(
                        "NULL, ", "")
                    if ref_corpus_id not in refs_done:
                        if ref in summary_quotes:
                            fixed_quote = summary_quotes[ref]["quote"]
                        else:
                            # abstract for inline citation
                            fixed_quote = inline_citation_quotes[ref]
                        fixed_quote = fixed_quote.strip().replace("“", '"').replace("”", '"')
                        if fixed_quote.startswith("..."):
                            fixed_quote = fixed_quote[3:]
                        if fixed_quote.endswith("..."):
                            fixed_quote = fixed_quote[:-3]
                        # dict to save reference strings as there is a possibility of having multiple papers in the same year from the same author
                        refs_done.add(ref_corpus_id)
                        ref_str_id = resolve_ref_id(ref_str, ref_corpus_id, citation_ids)
                        ref_data = pop_ref_data(ref_str_id, ref_corpus_id, fixed_quote,
                                                paper_metadata.get(ref_corpus_id))
                        if inline_tags:
                            curr_section["text"] = curr_section["text"].replace(ref, text_ref_format.format(
                                corpus_id=ref_data["paper"]["corpus_id"], ref_str=ref_data["id"]))
                        else:
                            curr_section["text"] = curr_section["text"].replace(ref, ref_data["id"])
                        refs_list.append(ref_data)
                else:
                    curr_section["text"] = curr_section["text"].replace(ref, "")
                    logger.warning(f"Reference not found in the summary quotes: {ref}")
            curr_section["text"] = re.sub(r"[ ]+", " ", curr_section["text"])
            # curr_section["text"] = curr_section["text"].replace(") ; (", "]; [")
            curr_section["citations"] = refs_list
            # add number of unique citations to section tldr
            if curr_section["tldr"]:
                if refs_list:
                    curr_section["tldr"] += (
                        f" ({len(refs_list)} sources)" if len(refs_list) > 1 else " (1 source)")
                else:
                    curr_section["tldr"] += " (LLM Memory)"
            sections.append(curr_section)
    return sections



from pydantic import BaseModel

from llms.constants import GPT_5_CHAT, GPT_4_1

OPENAI_API_KEY = "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"

import os

os.environ["OPENAI_API_KEY"] = OPENAI_API_KEY


class TaskResult(BaseModel):
    """The outcome of running a Task to completion"""
    name: str
    id: str
    roll_no: int
    class_: str


from litellm import completion

res = completion(messages=[{"role": "user", "content": "Create a random student record"}], model=GPT_4_1,
                 response_format=TaskResult)
print(res)



import logging
from scholarqa.llms.constants import *
from typing import List, Any, Callable, Tuple, Iterator, Union, Generator, Optional

import litellm
from litellm.caching import Cache
from litellm.utils import trim_messages
from langsmith import traceable

from scholarqa.state_mgmt.local_state_mgr import AbsStateMgrClient
from time import sleep

logger = logging.getLogger(__name__)


class CostAwareLLMCaller:
    def __init__(self, state_mgr: AbsStateMgrClient):
        self.state_mgr = state_mgr

    @staticmethod
    def parse_result_args(method_result: Union[Tuple[Any, CompletionResult], CompletionResult]) -> Tuple[
        Any, List[CompletionResult], List[str]]:
        if type(method_result) == tuple:
            result, completion_costs = method_result
        else:
            result, completion_costs = method_result, method_result
        completion_costs = [completion_costs] if type(completion_costs) != list else completion_costs
        completion_models = [cost.model for cost in completion_costs]
        return result, completion_costs, completion_models

    def parse_usage_args(self, method_result: Union[Tuple[float, TokenUsage], float]) -> Tuple[float, TokenUsage]:
        if isinstance(method_result, tuple):
            total_cost, tokens = method_result
        else:
            total_cost, tokens = method_result, TokenUsage(input=0, output=0, total=0, reasoning=0)
        return total_cost, tokens

    def call_method(self, cost_args: CostReportingArgs, method: Callable, **kwargs) -> CostAwareLLMResult:
        method_result = method(**kwargs)
        result, completion_costs, completion_models = self.parse_result_args(method_result)
        llm_usage = self.state_mgr.report_llm_usage(completion_costs=completion_costs, cost_args=cost_args)
        total_cost, tokens = self.parse_usage_args(llm_usage)
        return CostAwareLLMResult(result=result, tot_cost=total_cost, models=completion_models, tokens=tokens)

    def call_iter_method(self, cost_args: CostReportingArgs, gen_method: Callable, **kwargs) -> Generator[
        Any, None, CostAwareLLMResult]:
        all_results, all_completion_costs, all_completion_models = [], [], []
        for method_result in gen_method(**kwargs):
            result, completion_costs, completion_models = self.parse_result_args(method_result)
            all_completion_costs.extend(completion_costs)
            all_completion_models.extend(completion_models)
            all_results.append(result)
            yield result
        llm_usage = self.state_mgr.report_llm_usage(completion_costs=all_completion_costs, cost_args=cost_args)
        total_cost, tokens = self.parse_usage_args(llm_usage)
        return CostAwareLLMResult(result=all_results, tot_cost=total_cost, models=all_completion_models, tokens=tokens)


def success_callback(kwargs, completion_response, start_time, end_time):
    """required callback method to update the response object with cache hit/miss info"""
    completion_response.cache_hit = kwargs["cache_hit"] if kwargs["cache_hit"] is not None else False


NUM_RETRIES = 3
RETRY_STRATEGY = "exponential_backoff"
litellm.success_callback = [success_callback]


def setup_llm_cache(cache_type: str = "s3", **cache_args):
    logger.info("Setting up LLM cache...")
    litellm.cache = Cache(type=cache_type, **cache_args)
    litellm.enable_cache()


@traceable(run_type="llm", name="batch completion")
def batch_llm_completion(model: str, messages: List[str], system_prompt: str = None, fallback: Optional[str] = GPT_5_CHAT,
                         **llm_lite_params) -> List[Optional[CompletionResult]]:
    """returns the result from the llm chat completion api with cost and tokens used"""
    fallbacks = [f.strip() for f in fallback.split(",")] if fallback else []
    messages = [trim_messages([{"role": "system", "content": system_prompt}, {"role": "user", "content": msg}], model)
                for msg in messages]

    results, pending = [None] * len(messages), [_ for _ in range(len(messages))]
    curr_retry = 0

    #retries with exponential backoff in addition to fallbacks for pending instances
    while pending and curr_retry <= NUM_RETRIES:
        pending_msges = [messages[idx] for idx in pending]
        responses = litellm.completion_with_retries(messages=pending_msges, model=model, fallbacks=fallbacks,
                                                    retry_strategy=RETRY_STRATEGY, num_retries=NUM_RETRIES,
                                                    original_function=litellm.batch_completion, **llm_lite_params)

        for i, res in enumerate(responses):
            try:
                res_cost = round(litellm.completion_cost(res), 6)
                res_usage = res.usage
                reasoning_tokens = 0 if not (res_usage.completion_tokens_details and
                                             res_usage.completion_tokens_details.reasoning_tokens) else \
                    res_usage.completion_tokens_details.reasoning_tokens
                res_str = res["choices"][0]["message"]["content"].strip()
                cost_tuple = CompletionResult(content=res_str, model=res["model"],
                                              cost=res_cost if not res.get("cache_hit") else 0.0,
                                              input_tokens=res_usage.prompt_tokens,
                                              output_tokens=res_usage.completion_tokens,
                                              total_tokens=res_usage.total_tokens,
                                              reasoning_tokens=reasoning_tokens)
                results[i] = cost_tuple
            except Exception as e:
                if curr_retry == NUM_RETRIES:
                    logger.error(f"Error received for instance {i} in batch llm job, no more retries left: {e}")
                    raise e

        pending = [i for i, r in enumerate(results) if not r]
        curr_retry += 1
        if pending:
            logger.info(f"Retrying {len(pending)} failed instances in batch llm job, attempt {curr_retry}")
            sleep(2 ** curr_retry)

    return results


@traceable(run_type="llm", name="completion")
def llm_completion(user_prompt: str, system_prompt: str = None, fallback=GPT_5_CHAT, **llm_lite_params) -> CompletionResult:
    """returns the result from the llm chat completion api with cost and tokens used"""
    messages = []
    fallbacks = [f.strip() for f in fallback.split(",")] if fallback else []
    if system_prompt:
        messages.append({"role": "system", "content": system_prompt})
    messages.append({"role": "user", "content": user_prompt})
    response = litellm.completion_with_retries(messages=messages, retry_strategy=RETRY_STRATEGY,
                                               num_retries=NUM_RETRIES, fallbacks=fallbacks, **llm_lite_params)

    res_cost = round(litellm.completion_cost(response), 6)
    res_usage = response.usage
    reasoning_tokens = 0 if not (res_usage.completion_tokens_details and
                                 res_usage.completion_tokens_details.reasoning_tokens) else \
        res_usage.completion_tokens_details.reasoning_tokens
    res_str = response["choices"][0]["message"]["content"]
    if res_str is None:
        logger.warning("Content returned as None, checking for response in tool_calls...")
        res_str = response["choices"][0]["message"]["tool_calls"][0].function.arguments
    cost_tuple = CompletionResult(content=res_str.strip(), model=response.model,
                                  cost=res_cost if not response.get("cache_hit") else 0.0,
                                  input_tokens=res_usage.prompt_tokens,
                                  output_tokens=res_usage.completion_tokens, total_tokens=res_usage.total_tokens,
                                  reasoning_tokens=reasoning_tokens)
    return cost_tuple



from typing import Type

from nora_lib.tasks.models import AsyncTaskState, R
from nora_lib.tasks.state import StateManager
from filelock import FileLock


class LockedStateManager(StateManager):
    def __init__(self, task_state_class: Type[AsyncTaskState[R]], state_dir) -> None:
        super().__init__(task_state_class, state_dir)

    def read_state(self, task_id: str) -> AsyncTaskState[R]:
        lock = FileLock(f"{task_id}.lock")
        with lock:
            return super().read_state(task_id)

    def write_state(self, state: AsyncTaskState[R]) -> None:
        lock = FileLock(f"{state.task_id}.lock")
        with lock:
            super().write_state(state)
